from openpack.basepack import Part


class WorkbookPart(Part):
    content_type = (
        "application/vnd.openxmlformats-officedocument." "spreadsheetml.sheet.main+xml"
    )
    rel_type = (
        "http://schemas.openxmlformats.org/officeDocument/2006"
        "/relationships/officeDocument"
    )


class WorksheetPart(Part):
    content_type = (
        "application/vnd.openxmlformats-officedocument." "spreadsheetml.worksheet+xml"
    )

    rel_type = (
        "http://schemas.openxmlformats.org/officeDocument/2006/"
        "relationships/worksheet"
    )


class StylesPart(Part):
    content_type = (
        "application/vnd.openxmlformats-officedocument." "spreadsheetml.styles+xml"
    )

    rel_type = (
        "http://schemas.openxmlformats.org/officeDocument/2006/" "relationships/styles"
    )



import io
import cherrypy
import decimal

from xlsxcessive.xlsx import save
from xlsxcessive.workbook import Workbook


XLSX_CT = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
HOMEPAGE_HTML = """\
<html>
<head>
    <title>XlsXcessive - Generate Excel Spreadsheets Using Python</title>
    <style type="text/css">
        input {
            width:100%;
        }
        .installation-tip {
            font-size:10pt;
            font-style:italic;
        }
        #copy {
            background-color:#EEEEEE;
            height:100%;
        }
        #demo {
            margin-left:1%;
        }
    </style>
</head>
<body>
    <div id="copy" style="float:left;width:40%;">
        <img
            src="/static/logo.png"
            title="No graphic designers were harmed in the creation of this logo." />
        <blockquote>
        A cross-platform Python library for generating Excel 2007 compatible
        spreadsheets. A <em>simple</em> interface to an <em>excessive</em> file
        format.
        </blockquote>
        <p>
        XlsXcessive makes generating Excel compatible OOXML spreasheets easy.
        It supports common features like multiple cell data types, formulas,
        styles, merged cells and multiple worksheets.
        </p>
        <p>
        It is open source software developed at
        <a href="http://www.yougov.com/">YouGov</a> and licensed under the
        MIT license.
        </p>
        <p>
        Some basic documenation is in the
        <a href="https://github.com/jaraco/xlsxcessive">
            <code>project repo</code>
        </a>. Also see the
        <a href="https://github.com/jaraco/xlsxcessive/blob/master/sample.py">
            <code>sample.py</code>
        </a> file for an an example that excercises most of the current features
        of the library.
    </div>
    <div id="demo" style="float:left;width:59%;">
        <h2>Try It!</h2>
        <p>
        Enter some numbers, words and formulas below and
        export them as an Excel spreadsheet.
        </p>
        <form method="GET" action="demo">
            <table border="0" style="width:100%">
                <tr>
                    <td align="center" colspan="2">A</td>
                    <td align="center">B</td>
                    <td align="center">C</td>
                    <td align="center">D</td>
                </tr>
                <tr>
                    <td>1</td>
                    <td><input name="A1" type="text" value="2" /></td>
                    <td><input name="B1" type="text" value="Hello World!" /></td>
                    <td><input name="C1" type="text" /></td>
                    <td><input name="D1" type="text" /></td>
                </tr>
                <tr>
                    <td>2</td>
                    <td><input name="A2" type="text" value="2" /></td>
                    <td><input name="B2" type="text" /></td>
                    <td><input name="C2" type="text" /></td>
                    <td><input name="D2" type="text" /></td>
                </tr>
                <tr>
                    <td>3</td>
                    <td><input name="A3" type="text" value="=A1+A2" /></td>
                    <td><input name="B3" type="text" /></td>
                    <td><input name="C3" type="text" /></td>
                    <td><input name="D3" type="text" /></td>
                </tr>
                <tr>
                    <td>4</td>
                    <td><input name="A4" type="text" /></td>
                    <td><input name="B4" type="text" value="3.14" /></td>
                    <td><input name="C4" type="text" value="3" /></td>
                    <td><input name="D4" type="text" value="=B4*C4" /></td>
                </tr>
            </table>
            <p style="text-align:center;">
                <button type="submit">
                    <span style="font-size:20pt;">
                        Export as .xlsx
                    </span>
                </button>
            </p>
        </form>
        <h2>Get It!</h2>
        <p>Install with <code>pip</code></p>
        <pre>
        pip install XlsXcessive
        </pre>
    </div>
</body>
</html>
"""


class HomePage:
    exposed = True

    def GET(self):
        return HOMEPAGE_HTML


class Demo:
    exposed = True

    def GET(self, **cells):
        if cells:
            return self._generate_xlsx(cells)
        return ''

    def _generate_xlsx(self, cells):
        workbook = Workbook()
        sheet = workbook.new_sheet('Demo Sheet')
        for row in range(1, 5):
            for col in 'ABCD':
                ref = '%s%d' % (col, row)
                data = cells.get(ref, '').strip()
                if data:
                    value = self._infer_value(data, sheet)
                    sheet.cell(ref, value=value)
        out = io.BytesIO()
        save(workbook, 'demo.xlsx', out)
        headers = cherrypy.response.headers
        headers['Content-Length'] = out.tell()
        headers['Content-Disposition'] = 'attachment; filename=demo.xlsx'
        headers['Content-Type'] = XLSX_CT
        out.seek(0)
        return out

    def _infer_value(self, data, sheet):
        try:
            value = int(data)
        except ValueError:
            try:
                value = decimal.Decimal(data)
            except decimal.InvalidOperation:
                if data[0] == '=':
                    value = sheet.formula(data)
                else:
                    value = data
        return value


conf = {
    '/': {
        'request.dispatch': cherrypy.dispatch.MethodDispatcher(),
        'tools.decode.on': True,
    },
    '/static': {
        'tools.staticdir.on': True,
        'tools.staticdir.dir': '/home/christian/src/xlsxcessive/www/static',
    },
}


def main():
    homepage = HomePage()
    homepage.demo = Demo()

    cherrypy.quickstart(homepage, '/', conf)


__name__ == '__main__' and main()



"""Classes that represent parts of an OOXML Worksheet."""

import operator
import string
import datetime
import numbers

try:
    from functools import singledispatchmethod  # type: ignore
except ImportError:
    from singledispatchmethod import singledispatchmethod  # type: ignore

from xml.sax.saxutils import escape
from xlsxcessive import markup
from xlsxcessive.cache import CacheDecorator


class UnsupportedDateBase(Exception):
    pass


@CacheDecorator()
def _coords_to_a1_helper(coords):
    # the following closure was adapted from
    # http://stackoverflow.com/questions/22708/how-do-i-find-the-excel-column-name-that-corresponds-to-a-given-integer
    def num_to_a(n):
        n -= 1
        if n >= 0 and n < 26:
            return chr(65 + n)
        else:
            return num_to_a(n // 26) + num_to_a(n % 26 + 1)

    a1_col = num_to_a(coords[1] + 1)
    return "%s%d" % (a1_col, coords[0] + 1)


class Formula:
    def __init__(self, source, initial_value=None, shared=False, master=None):
        self.source = source
        self.initial_value = initial_value
        self.shared = shared
        self.master = master
        self.index = None
        self.refs = []
        self._ref_str = ''

    def share(self, cell):
        self.refs.append(cell.reference)
        if len(self.refs) == 1:
            # This is the first cell that this formula is being applied to.
            # Return it directly.
            return self

        # A new cell is referring to this formula. Return a shared version that
        # points to this one as the master formula.
        return Formula(None, shared=True, master=self)

    @property
    def _refs(self):
        if self.shared and not self.master and not self._ref_str:
            # sort alphabetically and then by length to ensure correct ordering
            sc = sorted(self.refs)
            sc = sorted(self.refs, key=len)
            low, high = sc[0], sc[-1]
            self._ref_str = "%s:%s" % (low, high)
        return self._ref_str

    def __str__(self):
        if self.master is not None:
            return '<f t="shared" si="%s" />' % self.master.index
        attrs = filter(
            None,
            [
                't="shared"' if self.shared else '',
                'ref="%s"' % self._refs if self._refs else '',
                'si="%d"' % self.master.index if self.master else '',
                'si="%d"' % self.index if (self.shared and not self.master) else '',
            ],
        )
        sattrs = " %s" % (" ".join(attrs)) if attrs else ''
        ival = '<v>%s</v>' % self.initial_value if self.initial_value else ''
        return '<f %s>%s</f>%s' % (sattrs, self.source, ival)


class Worksheet:
    """An OOXML Worksheet."""

    def __init__(self, workbook, name, sheet_id, relation_id):
        """Creates a new Worksheet. Semi-private class.

        Usually instantiated through a Workbook instance.

        Arguments
        ---------

         - workbook ...... An xlsxcessive.workbook.Workbook instance.
         - name .......... A string name for the worksheet.
         - sheet_id ...... An integer ID that is not shared with any other sheet
                           in the workbook.
         - relation_id ... A relationship ID string that is not shared with any
                           other sheet in the workbook.

        Fully functional worksheets should include real values for workbook,
        sheet_id and relation_id although you can pass None for those values
        and still get a Worksheet instance.
        """
        self.workbook = workbook
        self.name = name
        self.sheet_id = sheet_id
        self.relation_id = relation_id
        self.ref = markup.worksheet_ref % self.__dict__
        self.rows = []
        self.row_map = {}
        # Track formulas for sharing them amongst cells
        self.formulas = []
        # For settings that apply to entire columns
        self.cols = []

    def row(self, number):
        """Returns a Row. If the row doesn't exist, it is created."""
        if number in self.row_map:
            return self.row_map[number]
        row = Row(self, number)
        self.rows.append(row)
        self.row_map[number] = row
        return row

    def cell(self, *args, **params):
        """Creates and returns a new Cell for this Worksheet.

        Passes *args and **params to the Cell class constructor.
        """
        params['worksheet'] = self
        cell = Cell(*args, **params)
        rowidx = int(cell.coords[0])
        row = self.row(rowidx + 1)
        row.add_cell(cell)
        return cell

    def formula(self, *args, **params):
        """Creates and returns a new Formula for this Worksheet.

        Passes *args and **params to the Formula class constructor.
        """
        f = Formula(*args, **params)
        f.index = len(self.formulas)
        self.formulas.append(f)
        return f

    def col(self, *args, **params):
        """Creates and returns a new Column object for this Worksheet.

        Passes *args and **params to the Column class constructor.
        """
        c = Column(self, *args, **params)
        self.cols.append(c)
        return c

    def __str__(self):
        merges = []
        rows = []
        # Sort to put the rows and cells in the correct order - it
        # seems like this matters to Excel (though Open Office doesn't
        # care).
        self.rows.sort(key=operator.attrgetter('number'))
        for row in self.rows:
            # First sort the keys alphanumerically
            row.cells.sort(key=operator.attrgetter('reference'))
            # Then by length to get the correct sort order for A1 notation
            # where AA1 > Z1.
            row.cells.sort(key=lambda c: len(c.reference))
            rows.append(str(row))
            merges.extend(row.merge_cells)
        rows = ''.join(rows)
        if self.cols:
            cols_ = ''.join(str(col) for col in self.cols)
            cols = '<cols>%s</cols>' % cols_
        else:
            cols = ''
        if merges:
            merge_elems = []
            for merge_range in merges:
                merge_elems.append('<mergeCell ref="%s" />' % merge_range)
            merge_cells = '<mergeCells>%s</mergeCells>' % "".join(merge_elems)
        else:
            merge_cells = ''
        return markup.worksheet % {
            'rows': rows,
            'cols': cols,
            'merge_cells': merge_cells,
        }


class Row:
    def __init__(self, sheet, number):
        self.sheet = sheet
        self.number = number
        self.cells = []
        self.cell_map = {}

        # populated during rendering with references of merge cells
        self.merge_cells = []

    def cell(self, *args, **params):
        cell = Cell(*args, **params)
        if cell.reference in self.cell_map:
            return self.cell_map[cell.reference]
        cell.coords = (self.number - 1, len(self.cells))
        self.add_cell(cell)
        return cell

    def add_cell(self, cell):
        self.cells.append(cell)
        self.cell_map[cell.reference] = cell

    def __str__(self):
        cells = []
        for c in self.cells:
            cells.append(str(c))
            if c.merge_range:
                self.merge_cells.append(c.merge_range)
        cells = ''.join(cells)
        return '<row r="%s">%s</row>' % (self.number, cells)


class Column:
    __slots__ = 'width', 'number', 'best_fit', 'style'

    def __init__(self, worksheet, **params):
        for name, value in params.items():
            setattr(self, name, value)
        if not hasattr(self, 'number'):
            raise ValueError("One of number or index must be supplied.")

    @property
    def index(self):
        return self.number - 1

    @index.setter
    def index(self, value):
        self.number = value + 1

    def __str__(self):
        params = {}

        if getattr(self, 'width', None) is not None:
            params['width'] = self.width
            params['customWidth'] = 1

        if getattr(self, 'best_fit', None) is not None:
            params['bestFit'] = self.best_fit

        if getattr(self, 'style', None) is not None:
            params['style'] = self.style

        if not params:
            return ''

        params.update(self._colspec)
        attrs = ' '.join(
            '{key}="{value}"'.format(**vars()) for key, value in params.items()
        )

        return '<col ' + attrs + '/>'

    @property
    def _colspec(self):
        return dict(min=self.number, max=self.number)


class Cell:
    __slots__ = (
        '_reference',
        '_coords',
        'cell_type',
        '_value',
        '_is_date',
        '_is_datetime',
        '_is_time',
        'worksheet',
        'format',
        'merge_range',
    )

    def __init__(
        self, reference=None, value=None, coords=None, format=None, worksheet=None
    ):
        self._reference = reference.upper() if reference else reference
        self._coords = coords
        if not self._reference and self._coords:
            self._reference = self._coords_to_a1()
        self.cell_type = None
        self._is_date = False
        self._is_datetime = False
        self._is_time = False
        self.worksheet = worksheet
        self.value = value
        self.format = format
        self.merge_range = None

    @classmethod
    def from_reference(cls, ref):
        return cls(reference=ref)

    @classmethod
    def from_coords(cls, coords):
        return cls(coords=coords)

    def merge(self, other):
        self.merge_range = "%s:%s" % (self.reference, other.reference)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        return self._set_value(value)

    @singledispatchmethod
    def _set_value(self, value):
        raise ValueError("Unsupported cell value: %r" % value)

    @_set_value.register(numbers.Number)
    def _set_number(self, value):
        self.cell_type = "n"
        self._value = value

    @_set_value.register(datetime.datetime)
    def _set_datetime(self, value):
        self._is_datetime = True
        self.value = self._serialize_datetime(value)

    def _date_base(self):
        return 1904 if self.worksheet and self.worksheet.workbook.date1904 else 1900

    @_set_value.register(datetime.date)
    def _set_date(self, value):
        self._is_date = True
        self.value = self._serialize_date(value)

    @_set_value.register(datetime.time)
    def _set_time(self, value):
        self._is_time = True
        self.value = self._serialize_time(value)

    @_set_value.register(str)
    def _set_str(self, value):
        self.cell_type = "inlineStr"
        self._value = escape(value)

    @_set_value.register(bytes)
    def _set_bytes(self, value):
        self.cell_type = "inlineStr"
        self._value = value

    @_set_value.register(Formula)
    def _set_formula(self, value):
        self.cell_type = 'str'
        if value.shared:
            value = value.share(self)
        self._value = value

    @_set_value.register(type(None))
    def _set_none(self, value):
        self._value = value

    # Implementation of DATEVALUE to meet the requirements
    # described in 3.17.4.1 of the OOXML spec part 4
    #
    # For 1900 based sytems:
    #
    # DATEVALUE("01-Jan-1900") results in the serial value 1.0000000...
    # DATEVALUE("03-Feb-1910") results in the serial value 3687.0000000...
    # DATEVALUE("01-Feb-2006") results in the serial value 38749.0000000...
    # DATEVALUE("31-Dec-9999") results in the serial value 2958465.0000000...
    #
    # Furthermore:
    #
    # DATEVALUE("28-Feb-1900") results in 59
    # DATEVALUE("01-Mar-1900") results in 61
    #
    # For 1904 based systems:
    #
    # DATEVALUE("01-Jan-1904") results in the serial value 0.0000000...
    # DATEVALUE("03-Feb-1910") results in the serial value 2225.0000000...
    # DATEVALUE("01-Feb-2006") results in the serial value 37287.0000000...
    # DATEVALUE("31-Dec-9999") results in the serial value 2957003.0000000...
    #
    def _serialize_date(self, dateobj):
        base = self._date_base()
        if base == 1900:
            if dateobj < datetime.date(1900, 3, 1):
                delta = datetime.date(base, 1, 1) - datetime.timedelta(days=1)
            else:
                delta = datetime.date(base, 1, 1) - datetime.timedelta(days=2)
        elif base == 1904:
            delta = datetime.date(base, 1, 1)
        else:
            raise UnsupportedDateBase('Date base must be either 1900 or 1904')
        return (dateobj - delta).days

    # Implementation of TIMEVALUE
    #
    # see OOXML spec part 4: 3.17.4.2 Time Representation
    #
    # TIMEVALUE("00:00:00") results in the serial value 0.0000000...
    # TIMEVALUE("00:00:01") results in the serial value 0.0000115...
    # TIMEVALUE("10:05:54") results in the serial value 0.4207639...
    # TIMEVALUE("12:00:00") results in the serial value 0.5000000...
    # TIMEVALUE("23:59:59") results in the serial value 0.9999884...
    #
    def _serialize_time(self, timeobj):
        # calculate number of seconds since 00:00:00
        seconds = timeobj.second + timeobj.minute * 60 + timeobj.hour * 60 * 60
        return seconds / 86400

    # combination of DATEVALUE and TIMEVALUE
    def _serialize_datetime(self, datetimeobj, base=1900):
        date_float = float(self._serialize_date(datetimeobj.date(), base))
        time_float = self._serialize_time(datetimeobj.time())
        return date_float + time_float

    def _format_value(self):
        if self.cell_type == 'inlineStr':
            return "<is><t>%s</t></is>" % self.value
        elif self.cell_type == 'n':
            return "<v>%s</v>" % self.value
        elif self.cell_type == 'str':
            return str(self.value)

    def __str__(self):
        attrs = [
            'r="%s"' % self.reference,
            't="%s"' % self.cell_type,
        ]
        if self.format:
            attrs.append('s="%d"' % self.format.index)
        # if we don't have an explicit format and the
        # value is a date, datetime or time
        # then try to apply a default format to the cell
        elif self.worksheet:
            if self._is_date:
                idx = self.worksheet.workbook.stylesheet.default_date_format.index
                attrs.append('s="%d"' % idx)
            elif self._is_datetime:
                idx = self.worksheet.workbook.stylesheet.default_datetime_format.index
                attrs.append('s="%d"' % idx)
            elif self._is_time:
                idx = self.worksheet.workbook.stylesheet.default_time_format.index
                attrs.append('s="%d"' % idx)
        return '<c %s>%s</c>' % (" ".join(attrs), self._format_value())

    @property
    def reference(self):
        return self._reference

    class Coords:
        def __get__(self, instance, other):
            if instance._coords:
                return instance._coords
            if instance._reference:
                return instance._a1_to_coords()

        def __set__(self, instance, value):
            instance._coords = value

    coords = Coords()

    def _coords_to_a1(self):
        return _coords_to_a1_helper(self._coords)

    def _a1_to_coords(self):
        def _p():
            i = 0
            while True:
                yield 26 ** i
                i += 1

        row = int(''.join(filter(str.isdigit, self._reference))) - 1
        col_ref = ''.join(filter(str.isupper, self._reference))
        col = 0
        mod = 0
        for p, letter in zip(_p(), reversed(col_ref)):
            charval = string.ascii_uppercase.index(letter)
            col += p * (charval + mod)
            mod = 1
        return row, col



class CacheDecorator:
    def __init__(self):
        self.cache = {}
        self.func = None

    def cached_func(self, *args):
        if args not in self.cache:
            self.cache[args] = self.func(*args)
        return self.cache[args]

    def __call__(self, func):
        self.func = func
        return self.cached_func



"""Smoke tests for xlsxcessive."""

import io

from xlsxcessive import workbook, xlsx
from xlsxcessive.worksheet import Cell


class TestWhenSaving:
    def test_fairly_involved_save(self):
        wb = workbook.Workbook()

        sheet = wb.new_sheet('Test Sheet')

        # a shared font
        bigguy = wb.stylesheet.new_format()
        bigguy.font(size=24)

        # add a border
        bigguy.border(top="medium", bottom="medium")

        # set a specific number format
        bigguy.number_format('0.00')

        # another shared font style
        boldfont = wb.stylesheet.new_format()
        boldfont.font(bold=True)

        # the API supports adding rows
        row1 = sheet.row(1)

        # rows support adding cells - cells can currently store strings, numbers
        # and formulas.
        a1 = row1.cell("A1", "Hello, World!", format=boldfont)
        row1.cell("C1", 42.0, format=bigguy)

        # cells can be merged with other cells - there is no checking on invalid merges
        # though. merge at your own risk!
        a1.merge(Cell('B1'))

        # adding rows is easy
        row2 = sheet.row(2)
        row2.cell("B2", "Foo")
        row2.cell("C2", 1, format=bigguy)

        # formulas are written as strings and can have default values
        shared_formula = sheet.formula("SUM(C1, C2)", 43.0, shared=True)

        row3 = sheet.row(3)
        row3.cell("C3", shared_formula, format=bigguy)

        # you can work with cells directly on the sheet
        sheet.cell('D1', 12)
        sheet.cell('D2', 12)
        sheet.cell('D3', shared_formula)

        # and directly via row and column indicies
        sheet.cell(coords=(0, 4), value=40)
        sheet.cell(coords=(1, 4), value=2)
        sheet.cell(coords=(2, 4), value=shared_formula)

        # you can share a formula in a non-contiguous range of cells
        times_two = sheet.formula('PRODUCT(A4, 2)', shared=True)
        sheet.cell('A4', 12)
        sheet.cell('B4', times_two)
        sheet.cell('C4', 50)
        sheet.cell('D4', times_two)

        # iteratively adding data is easy now
        for rowidx in range(5, 10):
            for colidx in range(5, 11, 2):
                sheet.cell(coords=(rowidx, colidx), value=rowidx * colidx)

        # write unicode value
        sheet.cell('G2', value=u"43\u00b0")

        output = io.BytesIO()

        xlsx.save(wb, 'testwb.xlsx', output)

        # something should now be in the StringIO object
        assert output.getvalue()



import random

from xlsxcessive.worksheet import Worksheet


class TestAddingCellsToWorksheet:
    def setup_method(self, method):
        self.sheet = Worksheet(None, 'test', None, None)

    def _coords_to_a1(self, coords):
        def num_to_a(n):
            if n < 0:
                return ""
            if n == 0:
                return "A"
            return num_to_a(n // 26 - 1) + chr(n % 26 + 65)

        return "%s%d" % (num_to_a(coords[1]), coords[0] + 1)

    def test_cell_has_correct_reference_when_added_by_coords(self):
        # create a cell in the sixth row, second column
        cell = self.sheet.cell(coords=(5, 1))
        actual = cell.reference
        assert actual == "B6"

        # let's create more cells
        for row in [random.randint(0, 10000) for i in range(0, 10)]:
            for col in [random.randint(0, 1000000) for i in range(0, 5000)]:
                coords = (col, row)
                cell = self.sheet.cell(coords=coords)
                expected = self._coords_to_a1(coords)
                assert cell.reference == expected, 'Expected %s but got %s for %s' % (
                    expected,
                    cell.reference,
                    coords,
                )

    def test_creating_cell_creates_row_if_it_doesnt_exist(self):
        assert not self.sheet.rows
        self.sheet.cell('A1')
        assert self.sheet.rows


class TestCallingRowMethod:
    def setup_method(self, method):
        self.sheet = Worksheet(None, 'test', None, None)

    def test_creates_row_when_it_doesnt_exist(self):
        assert not self.sheet.rows
        row = self.sheet.row(4)
        assert row in self.sheet.rows

    def test_returns_existing_row_when_it_exists(self):
        r3 = self.sheet.row(3)
        assert r3 is self.sheet.row(3)

    def test_sets_the_row_number_to_the_requested_number(self):
        row = self.sheet.row(3)
        assert row.number == 3
        assert self.sheet.row_map[3] == row
        assert self.sheet.rows[0].number == 3



import datetime

from xlsxcessive.worksheet import Cell
from xlsxcessive.workbook import (
    Workbook,
)  # used for testing date conversions based to 1904

# the ADDRESS function in Excel is useful for getting correct values to work from
# for these tests.


class TestCellCoordToA1Conversion:
    def test_cell_in_row_0_col_0_is_A1(self):
        c = Cell(coords=(0, 0))
        actual = c.reference
        assert actual == 'A1', actual

    def test_cell_in_row_0_col_25_is_Z1(self):
        c = Cell(coords=(0, 25))
        actual = c.reference
        assert actual == 'Z1', actual

    def test_cell_in_row_0_col_26_is_AA1(self):
        c = Cell(coords=(0, 26))
        actual = c.reference
        assert actual == 'AA1', actual

    def test_cell_in_row_9_col_52_is_BA10(self):
        c = Cell(coords=(9, 52))
        assert c.reference == 'BA10'

    def test_cell_in_row_0_col_1299_is_AWZ1(self):
        c = Cell(coords=(0, 1299))
        actual = c.reference
        assert actual == 'AWZ1', actual

    def test_cell_in_row_0_col_676_is_ZA1(self):
        c = Cell(coords=(0, 676))
        actual = c.reference
        assert actual == 'ZA1', actual


class TestCellA1ToCoordConversion:
    def test_cell_A1_is_in_row_0_col_0(self):
        c = Cell(reference='A1')
        actual = c.coords
        assert actual == (0, 0)

    def test_cell_Z1_is_in_row_0_col_25(self):
        c = Cell(reference='Z1')
        assert c.coords == (0, 25)

    def test_cell_AA1_is_in_row_0_col_26(self):
        c = Cell(reference='AA1')
        actual = c.coords
        assert actual == (0, 26)

    def test_cell_BA10_is_in_row_9_col_52(self):
        c = Cell(reference='BA10')
        assert c.coords == (9, 52)

    def test_cell_AWZ1_is_in_row_0_col_1299(self):
        c = Cell(reference='AWZ1')
        actual = c.coords
        assert actual == (0, 1299)

    def test_cell_BFR1_is_in_row_0_col_1525(self):
        c = Cell(reference='BFR1')
        actual = c.coords
        assert actual == (0, 1525)

    def test_cell_ZA1_is_in_row_0_col_676(self):
        c = Cell('ZA1')
        actual = c.coords
        assert actual == (0, 676), actual


class TestCreatingCellsFromCoordinates:
    def test_sets_A1_reference(self):
        c = Cell.from_coords((0, 0))
        assert c.reference == 'A1'


class TestCreatingCellsFromA1References:
    def test_reference_A1_sets_coordinates_0_0(self):
        c = Cell.from_reference('A1')
        assert c.coords == (0, 0)


class TestCreatingCellsWithLowerCaseReferences:
    def test_the_reference_is_converted_to_uppercase(self):
        c = Cell.from_reference('a1')
        assert c.reference == 'A1'


class TestCellValues:
    def test_string_values_are_escaped(self):
        c = Cell('A1', value="AT&T")
        actual = c.value
        expected = "AT&amp;T"
        assert actual == expected

    def test_unicode_values_are_escaped(self):
        c = Cell('A1', value="43\u00b0")
        actual = c.value
        expected = "43\u00b0"
        assert actual == expected

    def test_already_encoded_strings_are_not_escaped(self):
        c = Cell('A1', value="43\xc2\xb0")
        actual = c.value
        expected = "43\xc2\xb0"
        assert actual == expected


# From Section 18.17.4.2 of the OOXML spec
# ----------------------------------------
# The time component of a serial value ranges in value from 0-0.99999999, and
# represents times from the instant starting 0:00:00 (12:00:00 AM) to the last
# instant of 23:59:59 (11:59:59 P.M.), respectively. Going forward in time, the
# time component of a serial value increases by 1/86,400 each second. [Note: As
# such, the time 12:00 has a serial value time component of 0.5. end note]

ONE_SEC = 1.0 / 86400
NOON = ONE_SEC * 43200
MIDNIGHT = ONE_SEC * 86400


class TestCellDateTime:
    # the following tests expect the date base to be 1/1/1900
    # (which is the default; 1/1/1904 has to be set in the
    # workbook)
    def test_date_conversion_1900_1_1(self):
        c = Cell('A1', value=datetime.date(1900, 1, 1))
        assert c.value == 1

    def test_date_conversion_1900_2_28(self):
        c = Cell('A1', value=datetime.date(1900, 2, 28))
        assert c.value == 59

    def test_date_conversion_1900_3_1(self):
        c = Cell('A1', value=datetime.date(1900, 3, 1))
        assert c.value == 61

    def test_date_conversion_1910_2_3(self):
        c = Cell('A1', value=datetime.date(1910, 2, 3))
        assert c.value == 3687

    def test_date_conversion_2006_2_1(self):
        c = Cell('A1', value=datetime.date(2006, 2, 1))
        assert c.value == 38749

    def test_date_conversion_9999_12_31(self):
        c = Cell('A1', value=datetime.date(9999, 12, 31))
        assert c.value == 2958465

    # the following tests expect the date base to be 1/1/1904
    # the datebase is set in the workbook, thus we need
    # to instantaite a Workbook object first
    def test_date_conversion_1904_1_1_date1904(self):
        wb = Workbook()
        wb.date1904 = True
        sheet = wb.new_sheet('Sheet 1')
        c = sheet.cell('A1', value=datetime.date(1904, 1, 1))
        assert c.value == 0

    def test_date_conversion_1910_2_3_date1904(self):
        wb = Workbook()
        wb.date1904 = True
        sheet = wb.new_sheet('Sheet 1')
        c = sheet.cell('A1', value=datetime.date(1910, 2, 3))
        assert c.value == 2225

    def test_date_conversion_2006_2_1_date1904(self):
        wb = Workbook()
        wb.date1904 = True
        sheet = wb.new_sheet('Sheet 1')
        c = sheet.cell('A1', value=datetime.date(2006, 2, 1))
        assert c.value == 37287

    def test_date_conversion_9999_12_31_date1904(self):
        wb = Workbook()
        wb.date1904 = True
        sheet = wb.new_sheet('Sheet 1')
        c = sheet.cell('A1', value=datetime.date(9999, 12, 31))
        assert c.value == 2957003

    # time conversion tests
    def test_time_conversion_00_00_00(self):
        c = Cell('A1', datetime.time(0, 0, 0))
        assert c.value == 0.0

    def test_time_conversion_00_00_01(self):
        c = Cell('A1', datetime.time(0, 0, 1))
        assert c.value == ONE_SEC

    def test_time_conversion_10_05_54(self):
        c = Cell('A1', datetime.time(10, 5, 54))
        assert c.value == ONE_SEC * 36354

    def test_time_conversion_12_00_00(self):
        c = Cell('A1', datetime.time(12, 0, 0))
        assert c.value == NOON

    def test_time_conversion_23_59_59(self):
        c = Cell('A1', datetime.time(23, 59, 59))
        assert c.value == MIDNIGHT - ONE_SEC



workbook = """\
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<workbook
    xmlns="http://schemas.openxmlformats.org/spreadsheetml/2006/main"
    xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships">
<workbookPr date1904="%(date1904)s"/>
  <sheets>
    %(sheets)s
  </sheets>
</workbook>
"""

worksheet = """\
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<worksheet
    xmlns="http://schemas.openxmlformats.org/spreadsheetml/2006/main"
    xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships">
  %(cols)s
  <sheetData>
    %(rows)s
  </sheetData>
  %(merge_cells)s
</worksheet>
"""

worksheet_ref = '<sheet name="%(name)s" sheetId="%(sheet_id)s" r:id="%(relation_id)s"/>'

stylesheet = """\
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<styleSheet xmlns="http://schemas.openxmlformats.org/spreadsheetml/2006/main"
    xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships">
%(numfmts)s
%(fonts)s
<fills count="1"><fill /></fills>
%(borders)s
<cellStyleXfs count="1"><xf /></cellStyleXfs>
%(formats)s
</styleSheet>
"""



from xlsxcessive.parts import WorkbookPart, WorksheetPart, StylesPart
from openpack.officepack import OfficePackage


def save(workbook, filename, stream=None):
    """Save the given workbook with the given filename.

    If stream is provided and is a file-like object the .xlsx data
    will be written there instead.
    """
    pack = OfficePackage()
    wbp = WorkbookPart(pack, '/workbook.xml', data=str(workbook))
    pack.add(wbp)
    pack.relate(wbp)

    stp = StylesPart(pack, '/styles.xml', data=str(workbook.stylesheet))
    pack.add(stp)
    wbp.relate(stp)

    for i, worksheet in enumerate(workbook.sheets):
        wid = i + 1
        wsp = WorksheetPart(pack, "/worksheet%d.xml" % wid, data=str(worksheet))
        pack.add(wsp)
        wbp.relate(wsp, id=worksheet.relation_id)
    pack.save(stream or filename)
